import math
import numpy as np  
import pandas as pd
import random
import copy
import matplotlib.pyplot as plt
from sklearn import mixture
from sklearn.model_selection import train_test_split
from scipy.stats import multivariate_normal
    
def check_array(X, name="X", dim=2): 
    if not (type(X)==np.ndarray) or len(X.shape)!=dim:
            raise ValueError(name+" should be a {:}-dimensional Numpy array.".format(dim))
            
def gmm_scores(X_train, X_val, k, covariance_type='full', reg_covar=1e-06, random_state=42):
    
    '''
    This function trains a GMM and evaluate it in a holdout set using the mean log_likelihood of samples
    
    Inputs: - X_train: training set; 
            - X_val: holdout set; 
            - k: number of GMM components;
            - covariance_type: covariance type (scikit-learn implementation);
            - random_state: seed.
    
    Output: - log_likelihood in the holdout set
    '''
    
    clf = mixture.GaussianMixture(n_components=k, covariance_type=covariance_type, reg_covar=reg_covar, random_state=random_state)
    clf.fit(X_train)
    return clf.score(X_val)

def get_gmm(X, y, y_cat=False, num_comps=[2,5,10,15,20], val_size=0.33, reg_covar=1e-06, random_state=42):
    
    '''
    This function trains a GMM and evaluate it in a holdout set using the mean log_likelihood of samples
    
    Inputs: - X: numpy array of features; 
            - y: numpy array of labels;
            - y_cat: if we should consider y as categorical;
            - num_comps: numbers of GMM components to be tested;
            - val_size: size of holdout set used to validate the GMMs numbers of components
            - random_state: seed.
    
    Output: - GMM ou dictionary of GMMs
    '''
    covariance_type='full'
    
    #Checking input format
    check_array(X, name="X", dim=2)
    check_array(y, name="y", dim=1)
    
    #Y categorical/or with few values
    if y_cat: 
        classes=list(set(y))
        gmm={}

        for c in classes:
            #Selecting number of components
            X_gmm_train, X_gmm_val, _, _=train_test_split(X[y==c], X[y==c], test_size=val_size, random_state=random_state)
            scores=np.array([gmm_scores(X_gmm_train, X_gmm_val, k, covariance_type=covariance_type, reg_covar=reg_covar,  random_state=random_state) for k in num_comps])
            k_star=num_comps[np.argmax(scores)]

            #Training GMMs
            gmm[c] = mixture.GaussianMixture(n_components=k_star, covariance_type=covariance_type, reg_covar=reg_covar, random_state=random_state)
            gmm[c].fit(X[y==c])
        
        return gmm #it is a dictionary of GMMs
    
    #Y continuous/or with many values
    else: 
        #Selecting number of components
        X_gmm_train, X_gmm_val, y_gmm_train, y_gmm_val = train_test_split(X, y, test_size=val_size, random_state=random_state)
        Z_gmm_train=np.hstack((y_gmm_train.reshape((-1,1)), X_gmm_train))
        Z_gmm_val=np.hstack((y_gmm_val.reshape((-1,1)), X_gmm_val))
        scores=np.array([gmm_scores(Z_gmm_train, Z_gmm_val, k, covariance_type=covariance_type, reg_covar=reg_covar, random_state=random_state) for k in num_comps])
        k_star=num_comps[np.argmax(scores)]
        
        #Training GMM
        Z = np.hstack((y.reshape((-1,1)),X))
        gmm = mixture.GaussianMixture(n_components=k_star, covariance_type=covariance_type, reg_covar=reg_covar, random_state=random_state)
        gmm.fit(Z)
        
        return gmm #it is a GMM



import math
import numpy as np  
import pandas as pd
import random
import copy
import matplotlib.pyplot as plt
from sklearn import mixture
from sklearn.model_selection import train_test_split
from scipy.stats import multivariate_normal

def check_array(X, name="X", dim=2): 
    if not (type(X)==np.ndarray) or len(X.shape)!=dim:
            raise ValueError(name+" should be a {:}-dimensional Numpy array.".format(dim))
            
def MI_gmm_reg(X, y, gmm, feat, eps=10**-50):
    
    '''
    This function calculates the mutual information between y and X in cases where we assume y continuous/with many values.
    
    Inputs: - X: numpy array of features; 
            - y: numpy array of labels;
            - gmm: GMM trained model;
            - feat: features indexes (feat);
            - eps: small value so we can avoid taking log of zero in some cases
    
    Output: - dictionary containing the estimate for the mutual information between y and X,
              and the standard deviation of measurements calculated from the samples.
    '''
    
    n, d = X.shape
    components=gmm.n_components
    Z=np.hstack((y.reshape((-1,1)),X))
    feat2=[0]+[f+1 for f in feat] #feat2 includes y in addition de X[:,feat]. PS: we sum 1 because when we train GMM, the first variable is always y

    ### Calculating log-likelihood with samples (x_i,y_i) 
    like=np.zeros(n)
    for c in range(components):
        like+=gmm.weights_[c]*multivariate_normal.pdf(Z[:,feat2], gmm.means_[c][feat2], gmm.covariances_[c][feat2][:,feat2])

    log_like_xy=np.log(like + eps)

    
    ### Calculating log-likelihood with samples (x_i) 
    like=np.zeros(n)
    for c in range(components):
        like+=gmm.weights_[c]*multivariate_normal.pdf(Z[:,feat2[1:]], gmm.means_[c][feat2[1:]], gmm.covariances_[c][feat2[1:]][:,feat2[1:]])

    log_like_x=np.log(like + eps)

    
    ### Calculating log-likelihood with samples (y_i) 
    like=np.zeros(n)
    for c in range(components):
        like+=gmm.weights_[c]*multivariate_normal.pdf(Z[:,0], gmm.means_[c][0], gmm.covariances_[c][0][0])

    log_like_y=np.log(like + eps)
     
    
    #Output
    m=np.mean(log_like_xy-log_like_x-log_like_y)
    s=np.std(log_like_xy-log_like_x-log_like_y)
    
    return {'mi':m, 'std':s}



def MI_gmm_class(X, y, gmm, feat, eps=10**-50):
    
    '''
    This function calculates the mutual information between y and X in cases where we assume y categorical/with few values.
    
    Inputs: - X: numpy array of features; 
            - y: numpy array of labels;
            - gmm: dict. of GMM trained models;
            - feat: features indexes (feat);
            - eps: small value so we can avoid taking log of zero in some cases
    
    Output: - dictionary containing the estimate for the mutual information between y and X,
              and the standard deviation of measurements calculated from the samples.
    '''
    
    
    n,d=X.shape
    classes=list(set(y))
    p={}

    ### Calculating log-likelihood with samples (y_i) 
    like=np.zeros(n)
    for c in classes:
        p[c]=np.mean(y==c)
        like[y==c]=p[c]

    log_like_y=np.log(like + eps)
    
    
    ### Calculating log-likelihood with samples (x_i,y_i) 
    like=np.zeros(n)
    for c in classes:
        #X|Y
        like_aux=np.zeros(n)
        for comp in range(gmm[c].n_components):
            like_aux[y==c]+=gmm[c].weights_[comp]*multivariate_normal.pdf(X[y==c][:,feat], gmm[c].means_[comp][feat], gmm[c].covariances_[comp][feat][:,feat])

        #(X,Y)
        like[y==c]=p[c]*like_aux[y==c] 
    log_like_xy=np.log(like + eps)

    
    ### Calculating log-likelihood with samples (x_i)
    like=np.zeros(n)
    for c in classes:
        #X|Y
        like_aux=np.zeros(n)
        for comp in range(gmm[c].n_components):
            like_aux+=gmm[c].weights_[comp]*multivariate_normal.pdf(X[:,feat], gmm[c].means_[comp][feat], gmm[c].covariances_[comp][feat][:,feat])

        #Marginalization of (X,Y)
        like+=p[c]*like_aux

    log_like_x=np.log(like + eps)
    
    
    #Output
    m=np.mean(log_like_xy-log_like_x-log_like_y)
    s=np.std(log_like_xy-log_like_x-log_like_y)
    
    return {'mi':m, 'std':s}



def MI(cand, posic, r, X, y, gmm, include_cand = True, eps=10**-50):
    
    '''
    This function is an intermediary function between the main class and the two functions that make the calculation of the
    mutual information. It basically decides which of the two functions to use and if we should do the forward or backward step.
    
    Inputs: - cand: position of the candidate variable to be chosen;
            - posic: list with positions of the selected variables so far;
            - r: round;
            - X: numpy array of features; 
            - y: numpy array of labels;
            - gmm: model or dict. of GMM(s);
            - include_cand: include or remove variables (forwar/backward)d;
            - eps: small value so we can avoid taking log of zero in some cases
            
    Output: - cand: position of the candidate variable to be chosen;
            - dic: dictionary containing the estimate for the mutual information between y and X,
              and the standard deviation of measurements calculated from the samples. 
    '''
        
    n,d=X.shape
    aux = copy.deepcopy(posic)
    if include_cand:
        aux[r] = cand
    else:
        aux.remove(cand)
    
    if type(gmm)==dict:
        dic=MI_gmm_class(X, y, gmm, aux, eps)
    else:
        dic=MI_gmm_reg(X, y, gmm, aux, eps) 
    
    return cand, dic



name="info_selection/info_selection"
__version__ = "1.0.0"

###
import math
import numpy as np  
import pandas as pd
import random
import copy
import matplotlib.pyplot as plt
from sklearn import mixture
from sklearn.model_selection import train_test_split
from scipy.stats import multivariate_normal
from .gmm import *
from .mi import *

def check_array(X, name="X", dim=2): 
    if not (type(X)==np.ndarray) or len(X.shape)!=dim:
            raise ValueError(name+" should be a {:}-dimensional Numpy array.".format(dim))
    
class SelectVars:
    
    '''
    This is the main class of the package.
    '''
    
    selection_mode = None
    gmm = None
    n = None
    
    def __init__(self, gmm, selection_mode = 'forward'):
        """
        Inputs: - gmm: model or dict. of GMM(s);
                - selection_mode: forward/backward algorithms.
        """
        
        if not selection_mode in ['forward', 'backward']:
            raise ValueError("Selection model should be either 'forward' or 'backward'.")
            
        self.selection_mode = selection_mode
        self.gmm=gmm
    
    def fit(self, X, y, verbose=True, eps=0):
        
        '''
        This function order the features according to their importance - from 
        most important to least important (forward) or from least important to most importante (backward).

        Inputs: - X: numpy array of features; 
                - y: numpy array of labels;
                - verbose: print or not to print!?
                - eps: small value so we can avoid taking log of zero in some cases 
        '''
        
        #Checking input format
        check_array(X, name="X", dim=2)
        check_array(y, name="y", dim=1)
        
        '''Creating some important objects'''
        self.n, self.d = X.shape
        include_var = self.selection_mode == 'forward'  #True if include or False if remove

        self.delta_list = [] # list with history of % changes of mutual info when adding/removing the best/worst variables
        self.mi_list = []  # list with mutual info history by adding/removing the best/worst variables
        self.stds_list = []  # list with stds history and that we will use to calculate the standard error of MIs
        self.feat_hist=[] # history of variables at each round
        lista = list(range(self.d))  # list with indexes of all variables
        
        '''Defining number of iterations and list of features we use in each iteration'''
        if verbose: print("Let's start...\n")
            
        #The 'posic' list starts empty if you include
        if include_var:
            posic = []  # lista com posições das variáveis selecionadas até o momento
            self.feat_hist.append(copy.deepcopy(posic))
            rounds = range(self.d) 

            self.mi_list.append(0)
            self.stds_list.append(0)
            self.delta_list.append(0)
            
        #The 'posic' list starts full if we take it out
        else:
            posic = copy.deepcopy(lista)
            self.feat_hist.append(copy.deepcopy(posic))
            rounds = range(self.d-1) 

            if type(self.gmm)==dict:
                dic=MI_gmm_class(X, y, self.gmm, posic, eps)
            else:
                dic=MI_gmm_reg(X, y, self.gmm, posic, eps) 

            self.mi_list.append(dic['mi'])
            self.stds_list.append(dic['std'])
            self.delta_list.append(0)
        
        if verbose: print("Round = {:3d}   |   Î = {:5.2f}   |   Δ%Î = {:5.2f}   |   Features={}".format(0, self.mi_list[-1], 0, posic)) 
        
        
        '''Calculating the Mutual Information (forward or backward fashion)'''
        for r in rounds: # "r" of rounds/repetitions
         
            if include_var:
                posic.append(None)
                
            #Calcula MI entre y e X[:,(posic, cand)] -> cand: variável candidata a ser selecionada
            outputs = [MI(cand, posic, r, X, y, self.gmm, include_var, eps) for cand in lista]
    
            #Escolhendo variável que traz maior retorno
            MI_best=-math.inf
            
            for out in outputs:
                
                cand, dic = out
                MI_current = dic['mi']
                
                if MI_current > MI_best:
                    MI_best = MI_current
                    std_best = dic['std']
                    best_index = cand
         
            #Δ%Î
            if r==0 and include_var:
                self.delta_list.append(0)
            else:
                self.delta_list.append(MI_best/self.mi_list[-1]-1)
                
            #Updating variable list        
            lista.remove(best_index)
            if include_var:
                posic[r] = best_index
            else:
                posic.remove(best_index)
    
            #Updating lists 
            self.mi_list.append(MI_best)
            self.stds_list.append(std_best)
            self.feat_hist.append(copy.deepcopy(posic))
            
            #Verbose
            if verbose: print("Round = {:3d}   |   Î = {:5.2f}   |   Δ%Î = {:5.2f}   |   Features={}".format(r+1, MI_best, self.delta_list[-1], posic))
            

        
    def get_info(self): 
        
        '''
        This function creates and outputs a Pandas DataFrame with the history of feature importance.
        '''
        
        dic={'rounds': range(0,len(self.mi_list)), 
             'mi_mean': self.mi_list, 
             'mi_error': [s/np.sqrt(self.n) for s in self.stds_list],
             'delta': self.delta_list,
             'features':self.feat_hist,
             'num_feat':[len(l) for l in self.feat_hist]}
        return pd.DataFrame(dic).loc[:,['rounds','mi_mean','mi_error','delta','num_feat','features']]
     
    def plot_delta(self): 
        
        '''
        This function plots the history of percentual changes in the mutual information.
        '''
        
        l=self.delta_list
        plt.plot(list(range(1,len(l))),l[1:])
        plt.axhline(y=0, color='r', linestyle='--')
        plt.xlabel("Rounds")
        plt.ylabel("Δ% Mutual Information")
        plt.show()
        
    def plot_mi(self): 
        
        '''
        This function plots the history of the mutual information.
        '''
        
        l,s=self.mi_list, self.stds_list
        plt.errorbar(list(range(len(l))), l, yerr=np.array(s)/np.sqrt(self.n)) #
        plt.axhline(y=0, color='g', linestyle='--')
        plt.xlabel("Rounds")
        plt.ylabel("Mutual Information")
        plt.show()
        
    def transform(self, X, rd): 
        
        '''
        This transforms X using the round 'rd'. Examine the history dataframe and plot before choosing 'rd'.
        '''
        
        return X[:,self.get_info().loc[rd,'features']]




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="infoselect",
    version="1.0.1",
    author="Felipe Maia Polo & Felipe Leno da Silva",
    author_email="felipemaiapolo@gmail.com, f.leno@usp.br",
    description="Mutual Information Based Feature Selection in Python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/felipemaiapolo/infoselect',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['scipy','numpy','pandas','sklearn','matplotlib'],
) 
